/* ListEditor.m
 * Written By:  Thomas Burkholder
 *
 * You may freely copy, distribute, and reuse the code in this example.
 * NeXT disclaims any warranty of any kind, expressed or  implied, as to its
 * fitness for any particular use.
 */

#import "ListEditor.h"
#import <objc/objc-runtime.h>
#import <objc/Object.h>
#import "../Ranker.subproj/Ranker.h"
#import "../Utilities.subproj/SortedList.h"
#import "../Utilities.subproj/ClassAgent.h"

static ListEditor *topListEditor;

#define STRFOROBJSEL(_a,_b) (NXAtom)[_a perform:(([_a respondsTo:sel_getUid(_b)])?sel_getUid(_b):@selector(name))]

@implementation ListEditor

+ topListEditor
{
	return topListEditor;
}

- initWith:anObject inDocument:aDocument
{
	char buf[MAXPATHLEN + 1];
	id bundle;
	
	[super init];
	bundle = [NXBundle bundleForClass:[self class]];
	[bundle getPath:buf forResource:"ListEditor" ofType:"nib"];
	[NXApp loadNibFile:buf owner:self withNames:NO fromZone:[self zone]];
	[listBrowser setMatrixClass:[Ranker class]];
	[listBrowser setTarget:self];
	[listBrowser setDoubleAction:@selector(editSelection:)];
	[listBrowser setDelegate:self];
	
	rootClassNode = [[SortedList alloc] init];
	[rootClassNode setAgent:[[ClassAgent alloc] init]];
	
	[rootClassNode addObject:[Object class]];
	[classBrowser setDelegate:rootClassNode];
	[window makeFirstResponder:classBrowser];
	document = aDocument;
	list = anObject;
	selection = [[List alloc] init];
	[self reset];
	return self;
}

- free
{
	if (topListEditor == self) topListEditor = nil;
	[window setDelegate:nil];
	[window close];
	[selection free];
	return [super free];
}

- (unsigned)selectionCount { return [selection count]; } 

- getSelectionInto:(List *)objectList 
{
	int	i, count = [selection count];
	[objectList empty];
	for (i = 0; i < count; i++) {
		[objectList addObject:[selection objectAt:i]];
	}
	return self;
}

- redrawSelection 
{
	int count = [selection count];
	[window disableFlushWindow];
	while (count--) [self resetObject:[selection objectAt:count]];
	[whatToDoMatrix setEnabled:![list count]];
	[window reenableFlushWindow];
	[window flushWindow];
	return self;
}

- document { return document; }
- editedObject { return list; }
- window { return window; }

- (BOOL) wantsSelection 
{ 
	return [selection count] != 0;
}

- selectObjects:(List *)objectList 
{
	int	i, count = [objectList count];
	[selection empty];
	for (i = 0; i < count; i++) {
		id	object = [objectList objectAt:i];
		if ([list indexOf:object] != NX_NOT_IN_LIST)
			[selection addObject:object];
	}
	[listBrowser loadColumnZero];
	[self redrawSelection];
	return self;
}

- close {
	if (topListEditor == self) topListEditor = nil;
	[self closeSubeditors];
	// following line required!
	[document editorDidClose:self for:list];
	[self free];
	return self;
}

- openSubeditorFor:anObject 
{
	NXRect winFrame;
	id editor = [document getEditor:NO for:anObject];
	BOOL existed = (editor!=nil);

	editor = [document getEditor:YES for:anObject];

	// stupid bug: activate is defined about four times.  Can't typecast
	// because we don't know what class document is.  So we use perform:
	// to get rid of the warning.
	[editor perform:@selector(activate)];
	if (!existed) {
		[window getFrame:&winFrame];
		[[editor window] moveTo:winFrame.origin.x + 30.0 :winFrame.origin.y + 30.0];
	}
	return editor;
}

- closeSubeditors
{
	int 	count = [list count];
	while (count--) {
		id editor = [document getEditor:NO for:[list objectAt:count]];
		if (editor) [editor close];
	}
	return self;
}

- resetObject:anObject 
{
	unsigned 	lastSel = -1, index = [list indexOf:anObject];
	const char *p;

	if (index != NX_NOT_IN_LIST) {
		id	matrix = [listBrowser matrixInColumn:0];
		id	cell = [matrix cellAt:index :0];
	
		[window disableFlushWindow];
		p = STRFOROBJSEL(anObject,[selField stringValue]);
		if (!p || !p[0])
			p = "<Null Pointer>";
		[cell setStringValue:p];
		[matrix drawCell:cell];
		if ([selection indexOf:anObject] != NX_NOT_IN_LIST) {
			[matrix setState:1 at:index :0];
			[matrix highlightCellAt:index :0 lit:1];
			lastSel = index;
		}
		
		// Need to put in code to scrollSelToVisible...
		if (lastSel != -1) [matrix scrollCellToVisible:lastSel :0];
		[window reenableFlushWindow];
		[window flushWindow];
	} 
	return self;
}

- orderFront { return [window orderFront:self]; }

- (BOOL)activate { 
	topListEditor = self;
	[window makeKeyAndOrderFront:self];
	[whatToDoMatrix setEnabled:![list count]];
	if ([list count]) {
		id l = [list objectAt:0];
		[whatToDoMatrix selectCellAt:0:((l == [l class])?0:1)];
	}
	[document setSelectionFrom:self];
	return YES;
}

- makeSelectionVisible:(BOOL)showIt { return self; }

- (NXAtom)acceptsTypeFrom:(const NXAtom *)typeList
{
	int i = 0;
	if (!typeList) return NULL;
	while (typeList[i]) {
		if (IBObjectPboardType == typeList[i]) return IBObjectPboardType;
		i++;
	}
	return NULL;
}

- (BOOL)copySelection
{
	if (![selection count]) return NO;
	else {
	const char 	*typeList[1];
	id		pasteBoard = [Pasteboard newName:NXGeneralPboard];
	typeList[0] = IBObjectPboardType;
	[pasteBoard declareTypes:typeList num:1 owner:nil];
	[document copyObjects:selection type:IBObjectPboardType inPasteboard:pasteBoard];
	}
	return YES;
}

- (BOOL)pasteInSelection 
{
	id	pasteBoard = [Pasteboard newName:NXGeneralPboard];
	id 	objectList = [document pasteType:IBObjectPboardType fromPasteboard:pasteBoard parent:list];
	int i, count = [objectList count];
	unsigned index;

	if (!count) return NO;
	index = [list indexOf:[selection lastObject]];
	if (index == NX_NOT_IN_LIST) index = [list count];
	else index++;
	[selection empty];
	[window disableFlushWindow];
	for (i = 0; i < count; i++) {
		id	object = [objectList objectAt:i];
		if (index < [list count])
			[list insertObject:object at:index];
		else {
			[list addObject:object];
			index++;
		}
	}
	[self selectObjects:objectList];
	[listBrowser display];
	[window reenableFlushWindow];
	[window flushWindow];
	[document setSelectionFrom:self];
	[objectList free];
	return YES;
}

- (BOOL)deleteSelection 
{
	int		i, count = [selection count];

	if (!count) return NO;
	[window disableFlushWindow];
	for (i = 0; i < count; i++) {
		id	entry = [selection objectAt:i];
		[list removeObject:entry];
	}
	[listBrowser display];
	[document deleteObjects:selection];
	[selection empty];
	[listBrowser loadColumnZero];
	[self redrawSelection];
	[window reenableFlushWindow];
	[window flushWindow];
	[document setSelectionFrom:self];
	return YES;
}

- (int)browser:sender fillMatrix:matrix inColumn:(int)column 
{
	const char *p;

	int	i, count = [list count];
	[matrix renewRows:count cols:1];
	for (i = 0; i < count; i++) {
		id cell = [matrix cellAt:i :0];
		p = STRFOROBJSEL([list objectAt:i],[selField stringValue]);
		if (!p || !p[0])
			p = "<Null Pointer>";
		[cell setStringValue:p];
		[cell setLeaf:YES];
		[cell setLoaded:YES];
	}
	return count;
}

- reset 
{
	[listBrowser loadColumnZero];
	[classBrowser reloadColumn:[classBrowser lastColumn]];
	[self setSelection:self];
	return self;
}

- setSelection:sender
{
	id	selectedCells = [List new], matrix = [listBrowser matrixInColumn:0];
	id	cellList = [matrix cellList];
	int	count;
	[NXApp stopConnecting];
	[matrix getSelectedCells:selectedCells];
	[selection empty];
	count = [selectedCells count];
	while (count--) {
		[selection addObject:[list objectAt:[cellList indexOf:[selectedCells objectAt:count]]]];
	}
	[document setSelectionFrom:self];
	[selectedCells free];
	return self;
}

- editSelection:sender
{
	if ([selection count] != 1)
		NXBeep();
	else if (![self openSubeditorFor:[selection objectAt:0]])
		NXBeep();
	return self;
}

- windowWillClose:sender 
{
	[self close];
	return nil; // window has been already closed by close
}

- windowDidBecomeKey:sender 
{
	[self activate];
	return self;
}

- addEntry:entry
{
	unsigned 	index = [list indexOf:[selection lastObject]];
	List	*aList = [List new];

	if (index == NX_NOT_IN_LIST) index = [list count];
	else index++;
	[selection empty];
	if (index < [list count])
		[list insertObject:entry at:index];
	else
		[list addObject:entry];
	[aList addObject:entry];
	[self selectObjects:aList];
	[aList free];
	return self;
}

- add:sender
{
	id obj;

	obj = objc_lookUpClass([[classBrowser selectedCell] stringValue]);
	if ([whatToDoMatrix selectedCol]) {
		return [self addEntry:[[obj alloc] init]];
	}else if (obj && [obj instancesRespondTo:@selector(init)]) {
		return [self addEntry:obj];
	} else {
		NXBeep();
		return nil;
	}
}

- remove:sender
{
	BOOL yn;

	yn = [self deleteSelection];
	if (!yn) NXBeep();
	return self;
}

- ok:sender
{
	[self reset];
	return self;
}

@end

@implementation List (IBEditorMethods)

- (const char *)getEditorClassName
{
	NXEvent *ev;

	ev = [NXApp currentEvent];
	if (ev->flags&NX_ALTERNATEMASK) // the superclass's editor
		return [super getEditorClassName];
	else // the real inspector
		return "ListEditor";
}

- (NXImage *)getIBImage
{
	return [NXImage findImageNamed:"ListIcon"];
}

@end

